//------------------------------------------------------------------------------
//
// app.h
//
//------------------------------------------------------------------------------
#include "common.h"

//internal
HRESULT InitDirectX();
HRESULT TermDirectX();

HRESULT InitDirectInput();
HRESULT TermDirectInput();

HRESULT InitDirectSound();
HRESULT TermDirectSound();

static HWND      hWnd          = NULL;
static HINSTANCE hInst         = NULL;
static DWORD     dwWidth       = 0;
static DWORD     dwHeight      = 0;
static BOOL      bIsWindow     = FALSE;
static char      szAppName[MAX_PATH + 1];

static WORD      wMouseX       = 0;
static WORD      wMouseY       = 0;
static BOOL      bMouseLButton = FALSE;
static BOOL      bMouseRButton = FALSE;
static BOOL      bMouseMButton = FALSE;

void ShowErrorMessage(LPCSTR message, LPCSTR caption) {
  MessageBox(NULL, message, caption, MB_OK);
}

POINT GetMousePos() {
  POINT p = {0, 0};
  if(hWnd) {
    GetCursorPos(&p);
    ScreenToClient(hWnd, &p);
  }
  return p;
}

DWORD GetWidth() {
  return dwWidth;
}

DWORD GetHeight() {
  return dwHeight;
}

BOOL GetIsWindow() {
  return bIsWindow;
}

HWND GetWindowHandle() {
  return hWnd;
}

HINSTANCE GetInstanceHandle() {
  return hInst;
}

static LRESULT WINAPI SimpleMsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
  switch( msg ) {
    //need 
    case WM_SYSCOMMAND:
      switch( ( wParam & 0xFFF0 ) ) {
        case SC_MONITORPOWER:
        case SC_SCREENSAVE:
          return 0; // DefWindowProc not call
          break;
      }
      break;

    case WM_PAINT:
      break;

    case WM_CLOSE:
    case WM_DESTROY:
      ::PostQuitMessage( 0 );
      return 0;
      break;

    case WM_KEYDOWN:
      if( wParam == VK_ESCAPE ) {
        ::PostQuitMessage( 0 );
      }
      break;

    case WM_LBUTTONDOWN:
      bMouseLButton = TRUE;
      break;
      
    case WM_RBUTTONDOWN:
      bMouseRButton = TRUE;
      break;
      
    case WM_MBUTTONDOWN:
      bMouseMButton = TRUE;
      break;

    case WM_LBUTTONUP:
      bMouseLButton = FALSE;
      break;
      
    case WM_RBUTTONUP:
      bMouseRButton = FALSE;
      break;
      
    case WM_MBUTTONUP:
      bMouseMButton = FALSE;
      break;
    
     
    case WM_IME_SETCONTEXT:
      lParam &= ~ISC_SHOWUIALL;
      break;
  }
  return ::DefWindowProc( hWnd, msg, wParam, lParam );
}

HRESULT Term() {
  TermDirectSound();
  TermDirectInput();
  TermDirectX();
  if(hInst) {
    ::DestroyWindow(hWnd);
    ::timeEndPeriod(1);
    ::UnregisterClass(szAppName, hInst);
    hInst = NULL;
    hWnd  = NULL;
  }
  return S_OK;
}

//entry point
HRESULT Init(LPCSTR szName, DWORD w, DWORD h, BOOL bWindow) {
  HRESULT   hRet        = E_FAIL;
  DWORD     style       = 0;
  DWORD     dwExStyle   = 0;
  POINT     point;
  RECT      rc;
  hInst     = ::GetModuleHandle(NULL);
  dwWidth   = w;
  dwHeight  = h;
  bIsWindow = bWindow;
  wsprintf(szAppName, "%s", szName);
  
  WNDCLASSEX twc = {
    sizeof( WNDCLASSEX ), CS_CLASSDC | CS_VREDRAW | CS_HREDRAW, SimpleMsgProc, 0L, 0L,
    hInst, ::LoadIcon( hInst, TEXT( "MAINICO" ) ), ::LoadCursor(NULL, IDC_ARROW), 
    ( HBRUSH )::GetStockObject(BLACK_BRUSH), NULL, szName, NULL
  };
  
  if ( !::RegisterClassEx( &twc ) ) {
    ShowErrorMessage("RegisterClassEx : Failed\n", "Error");
    Term();
    return hRet;
  }
  
  if(!bWindow) {
    dwExStyle = WS_EX_APPWINDOW;
    style     = WS_POPUP;
  } else {
    dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
    style     = WS_OVERLAPPEDWINDOW;
  }
  
  ::SetRect( &rc, 0, 0, dwWidth, dwHeight );
  point.x     = ::GetSystemMetrics( SM_CXSCREEN );
  point.y     = ::GetSystemMetrics( SM_CYSCREEN );
  ::AdjustWindowRectEx( &rc, style, FALSE, dwExStyle);

  hWnd = ::CreateWindowEx(
    dwExStyle, szAppName, szAppName, style,
    ( point.x - dwWidth ) / 2, ( point.y - dwHeight ) / 2,
    rc.right - rc.left, rc.bottom - rc.top,
    NULL, NULL, hInst, NULL );
  
  if(!hWnd) {
    ShowErrorMessage("CreateWindowEx : Failed\n", "Error");
    Term();
    return hRet;
  }
  
  //Show Window
  ::ShowWindow(hWnd, SW_SHOW);
  ::UpdateWindow(hWnd);
  ::timeBeginPeriod(1);
  
  //Initialize DirectInput
  hRet = InitDirectInput();
  if(hRet != DI_OK) {
    ShowErrorMessage("InitDirectInput : Failed\n", "Error");
    Term();
    return hRet;
  }
  
  //Initialize DirectSound
  hRet = InitDirectSound();
  if(hRet != D3D_OK) {
    ShowErrorMessage("InitSound : Failed\n", "Error");
    Term();
    return hRet;
  }
  
  //Initialize DirectGraphics
  hRet = InitDirectX();
  if(hRet != D3D_OK) {
    ShowErrorMessage("InitDirectX : Failed\n", "Error");
    Term();
    return hRet;
  }
  
  return hRet;
}


BOOL DoEvent() {
  MSG   msg;
  BOOL  bActive = TRUE;
  if(hWnd) {
    while(::PeekMessage(&msg , 0 , 0 , 0 , PM_REMOVE)) {
      if(msg.message == WM_QUIT) {
        bActive = FALSE;
        break;
      } else {
        ::TranslateMessage( &msg );
        ::DispatchMessage( &msg );
      }
    }
    Sleep(0);
  }
  
  //Update Key
  UpdateKeyState();
  
  return bActive;
}

void Quit() {
  ::PostQuitMessage( 0 );
}

