#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cell.h"

static cell s_nil;
static cell *nil = &s_nil;
static cell *root = nil;
static cell *empty = nil;
static cell pool[CELL_MAX];

void init_cell() {
  nil = &s_nil;
  nil->first = nil;
  nil->rest = nil;
  nil->mark = 0;
  for(int i = 0 ; i < CELL_MAX - 1; i++) {
    pool[i].rest = &(pool[i + 1]);
  }
  pool[CELL_MAX - 1].rest = nil;
  
  empty = pool;
  root = nil;
}

int is_nil_cell(cell *t) {
  if(nil == t) return 1;
  return 0;
}

int is_gc_cell(cell *t) {
  if(t && t != nil) {
    return t->mark & cell::gc;
  }
  return 0;
}

cell *nil_cell() {
  return nil;
}

void update_dummy_cell(cell *) {
}

void render_dummy_cell(cell *) {
}

cell *add_cell( void (*update)(cell *), void (*render)(cell *) ) {
  cell *t = empty;
  if(t != nil) {
    empty = t->rest;
    t->mark = 0;
    t->first = nil;
    t->rest = root;
    root = t;
    t->count = 0;
    
    if(update) t->update = update;
    else       t->update = update_dummy_cell;
    
    if(render) t->render = render;
    else       t->render = render_dummy_cell;
  }
  return t;
}

__forceinline void func_update_cell(cell *t) {
  t->update(t);
  t->count++;
}

__forceinline void func_render_cell(cell *t) {
  t->render(t);
}

__forceinline int do_function_sub_cell( cell *t, void (*func)(cell *t), int count) {
  if(t == nil) return count;
  count++;
  func(t);
  if(t->first != nil) {
    do_function_sub_cell(t->first, func, count);
  }
  return do_function_sub_cell(t->rest, func, count);
}


__forceinline int do_update_sub_cell(cell *t) {
  int ret = 0;
  while( !is_nil_cell(t) ) {
    if( !is_nil_cell(t->first) ) {
      ret += do_update_sub_cell(t->first);
    }
    t->update(t);
    t->count++;
    t = t->rest;
    ret++;
  }
  return ret;
}



int do_update_cell_r() {
  return do_function_sub_cell(root, func_update_cell, 0);
}

int do_update_cell() {
  return do_update_sub_cell(root);
}

int do_render_sub_cell(cell *t) {
  int ret = 0;
  while(t != nil) {
    if(t->first != nil) {
      ret += do_render_sub_cell(t->first);
    }
    t->render(t);
    t = t->rest;
    ret++;
  }
  return ret;
}

int do_render_cell() {
  return do_render_sub_cell(root);
}

__forceinline cell *do_gc_sub_cell(cell *t) {
  if(is_nil_cell(t)) return nil_cell();
  if( !is_gc_cell(t) ) {
    t->rest = do_gc_sub_cell(t->rest);
    return t;
  }
  cell *rest = t->rest;
  t->rest = empty;
  empty = t;
  return do_gc_sub_cell(rest);
}

cell *do_gc_cell() {
  return root = do_gc_sub_cell(root);
}

cell *mark_gc_cell(cell *t) {
  if(t && t != nil) {
    t->mark |= cell::gc;
  }
  return t;
}

int count_cell() {
  int ret = 0;
  cell *t = root;
  while(t != nil) {
    ret++;
    t = t->rest;
  }
  return ret;
}


#ifdef _CELL_TEST_

void test_update(cell *t)  {
  printf("%08X: update\n");
}

void test_render(cell *t)  {
  printf("%08X: render\n");
}


int main () {
  init_cell();
  cell *t
  while( is_nil_cell(t = add_cell(test_update, test_render) ) );

  do_update_cell();
  do_render_cell();
  return 0;
}

#endif

