//------------------------------------------------------------------------------
//
// common.h
//
//------------------------------------------------------------------------------
#ifndef _COMMON_H_
#define _COMMON_H_

//------------------------------------------------------------------------------
// INCLUDE
//------------------------------------------------------------------------------
//windows
#include <windows.h>

//std
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

//DirectX
#include <d3d9.h>
#include <d3dx9.h>
#include <dsound.h>

#define DIRECTINPUT_VERSION    0x800
#include <dinput.h>

//------------------------------------------------------------------------------
// MACRO
//------------------------------------------------------------------------------
//COM
#define RELEASE(x)                            { if( (x) ) { x->Release(); } x = NULL; }
#define SAFE_DELETE(x)                        { if( (x) ) { delete     x; } x = NULL; }
#define SAFE_DELETE_ARRAY(x)                  { if( (x) ) { delete []  x; } x = NULL; }

//D3DXMisc
#define RAD(x)                                      D3DXToRadian(x)
#define PI                                          D3DX_PI

//Vector2
#define Vector2                                     D3DXVECTOR2
#define Vector2Trans2D(dest, src, m)                D3DXVec2TransformCoord(dest, src, m)
#define Vector2Length(v)                            D3DXVec2Length(v)
#define Vector2Normalize(out, in)                   D3DXVec2Normalize(out, in)
#define Vector2Direction(out, from, to)             { out = to - from; Vector2Normalize(&out, &out); }
//Vector3
#define Vector3                                     D3DXVECTOR3
#define Vector3Normalize(out, in)                   D3DXVec3Normalize(out, in)
#define Vector3Dot(v1, v2)                          D3DXVec3Dot(v1, v2)
#define Vector3Length(v)                            D3DXVec3Length(v)
#define Vector3Len(v)                               Vector3Length(v)
#define Vector3Cross(out, v1, v2)                   D3DXVec3Cross(out, v1, v2)
#define Vector3Lerp(out, v1, v2, t)                 D3DXVec3Lerp(out, v1, v2, t)
#define Vector3Direction(out, from, to)             { out = to - from; Vector3Normalize(&out, &out); }

//Vector4
#define Vector4                                     D3DXVECTOR4

//Matrix
#define Matrix                                      D3DXMATRIX
#define MatrixTrans(out, x, y, z)                   D3DXMatrixTranslation(out, x, y, z)
#define MatrixView(out, pos, at, up)                D3DXMatrixLookAtLH(out, pos, at, up)
#define MatrixProj(out, fov, aspect, zn, zf)        D3DXMatrixPerspectiveFovLH(out, fov, aspect, zn, zf)
#define MatrixProjection(out, fov, aspect, zn, zf)  MatrixProj(out, fov, aspect, zn, zf)
#define MatrixTranspose(mout, min)                  D3DXMatrixTranspose(mout, min)
#define MatrixRotateX(m, angle)                     D3DXMatrixRotationX(m, angle)
#define MatrixRotateY(m, angle)                     D3DXMatrixRotationY(m, angle)
#define MatrixRotateZ(m, angle)                     D3DXMatrixRotationZ(m, angle)
#define MatrixRotateAxis(out, vec, fa)              D3DXMatrixRotationAxis(out, vec, fa)
#define MatrixScalling(out, x, y, z)                D3DXMatrixScaling(out, x, y, z)

//Color
#define Color                                       D3DCOLOR
#define ColorFloat                                  D3DXCOLOR

//#define DefaultAspect                             ((FLOAT)GetWidth() / (FLOAT)GetHeight())
#define DefaultFov                                  (D3DXToRadian(30))
#define DefaultAspect                               ( static_cast<FLOAT>(GetWidth()) / static_cast<FLOAT>(GetHeight()) )
#define DefaultZNear                                (0.5f)
#define DefaultZFar                                 (1000.0f)
#define DefaultViewPos                              ( D3DXVECTOR3(0.0f, 10.0f, -10.0f)  )
#define DefaultViewAt                               ( D3DXVECTOR3(0.0f,  0.0f,  0.0f)   )
#define DefaultViewUp                               ( D3DXVECTOR3(0.0f,  1.0f,  0.0f)   )
#define DefaultLightAmbient                         (0x11111111)
#define DefaultLightColor                           ( D3DXCOLOR(0.5f, 0.5f, 0.5f, 0.5f) )
#define DefaultLightVector                          ( -DefaultViewPos  )

//Saturate
#define Sat8Bit(a)  ( (a > 255)   ?   255 : ((a < 0) ? 0 : a) )
#define Sat16Bit(a) ( (a > 32767) ? 32767 : ((a < 0) ? 0 : a) )

//------------------------------------------------------------------------------
// ENUMERATE
//------------------------------------------------------------------------------
enum {
  LostSleepTime   = 250,
  MaxLostRepeat   = 30,
  FlipTextureMax  = 2,
  FontMax         = 32,
  ShaderMax       = 32,
  TextureMax      = 256,
  MeshMax         = 256,
  SoundMax        = 256,
  PollThreadSleep = 500,
};

enum {
  BlendNormal  = 0x00000001,
  BlendAdd     = 0x00000002,
  TssNormal    = 0x00000010,
  FilterPoint  = 0x00000100,
  FilterLinear = 0x00000200,
  ZBufferOn    = 0x00000400,
  FogOn        = 0x00000800,
  CullOn       = 0x00001000,
  LightOn      = 0x00002000,
};

enum {
  Keyboard = 0,
  Joystick,
  Max,
};

enum {
  CoopLevel = DISCL_FOREGROUND | DISCL_NONEXCLUSIVE,
}; 

enum {
  Up            = 0x00000001,
  Down          = 0x00000002,
  Left          = 0x00000004,
  Right         = 0x00000008,
  Button_0      = 0x00000100,
  Button_1      = 0x00000200,
  Button_2      = 0x00000400,
  Button_3      = 0x00000800,
  Button_4      = 0x00001000,
  Button_5      = 0x00002000,
  Button_6      = 0x00004000,
  Button_7      = 0x00008000,
  Button_8      = 0x00010000,
  Button_9      = 0x00020000,
  Pushed        = 0x0003ffff,
  PushedButton  = 0x0003ff00,
};

//------------------------------------------------------------------------------
// Windows API
//------------------------------------------------------------------------------
void    ShowErrorMessage(LPCSTR caption, LPCSTR message);
POINT   GetMousePos();
DWORD   GetWidth();
DWORD   GetHeight();
BOOL    GetIsWindow();
HWND    GetWindowHandle();
HINSTANCE GetInstanceHandle();
HRESULT Init(LPCSTR szName, DWORD w, DWORD h, BOOL bWindow);
HRESULT Term();
BOOL    DoEvent();
void    Quit();

//------------------------------------------------------------------------------
// DirectX Graphics API
//------------------------------------------------------------------------------
HRESULT BeginScene();
HRESULT EndScene();
HRESULT Flip(D3DCOLOR = 0xFFFFFFFF);
HRESULT SetBlur(D3DCOLOR dwColor);
HRESULT EnableBlur();
HRESULT DisableBlur();
HRESULT Clear(D3DCOLOR color, BOOL bIsAlpha = FALSE);
HRESULT StartScene(D3DCOLOR color, BOOL bIsAlpha = FALSE);
LPCSTR  DisplayDriverName();
LPCSTR  DisplayDescriptionName();
LPDIRECT3DDEVICE9 GetDevice();

inline HRESULT EnableMultiSample() {
  return GetDevice()->SetRenderState(D3DRS_MULTISAMPLEANTIALIAS, TRUE);
}

inline HRESULT DisableMultiSample() {
  return GetDevice()->SetRenderState(D3DRS_MULTISAMPLEANTIALIAS, FALSE);
}

inline HRESULT SetShadeNormal() {
  return GetDevice()->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_GOURAUD);
}

inline HRESULT SetShadeFlat() {
  return GetDevice()->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_FLAT);
}

//Texture Load API
HRESULT LoadTextureFromFile(LPCSTR szName);
HRESULT UnloadTexture(LPCSTR szName);

//Setting API
HRESULT SetRenderState(DWORD dwBits);
DWORD   GetRecordRenderState();
HRESULT SetTexture(LPCSTR szName);
HRESULT SetFogParam(FLOAT fStart, FLOAT fEnd, D3DCOLOR dwColor);
HRESULT SetProjection(FLOAT fFov, FLOAT fAspect, FLOAT fNear, FLOAT fFar);
HRESULT SetProjection(D3DXMATRIX &mProj);
HRESULT SetView(D3DXVECTOR3 &vPos, D3DXVECTOR3 &vAt, D3DXVECTOR3 &vUp);
HRESULT SetView(D3DXMATRIX &mView);
HRESULT SetWorld(D3DXMATRIX &mWorld);
HRESULT SetLight(BOOL b, D3DCOLOR ambient = DefaultLightAmbient, D3DXCOLOR col = DefaultLightColor, D3DXVECTOR3 &v = DefaultLightVector);
HRESULT SetAmbient(D3DCOLOR amb);

//Font API
HRESULT CreateFontData(DWORD dwNum, LPCSTR szName, UINT pitch);
HRESULT ReleaseFontData(DWORD dwNum);
HRESULT DrawFontData(DWORD dwNum, FLOAT x, FLOAT y, D3DCOLOR color, LONG shade, const char *fmt, ...);

//Draw API
HRESULT DrawRect2D(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2, D3DCOLOR color, FLOAT u1, FLOAT v1, FLOAT u2, FLOAT v2);
HRESULT DrawRect2D(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2, D3DCOLOR color);
HRESULT DrawRectCenter2D(FLOAT x, FLOAT y, FLOAT w, FLOAT h, D3DCOLOR color);
HRESULT DrawLine2D(D3DXVECTOR2 &p1, D3DXVECTOR2 &p2, D3DCOLOR color1, D3DCOLOR Color2);

//Batch Draw API
HRESULT DrawBatchRect2D(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2, D3DCOLOR color);
HRESULT DrawBatchRectCenter2D(FLOAT x, FLOAT y, FLOAT w, FLOAT h, D3DCOLOR color);
HRESULT DrawBatchRect2D(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2, D3DCOLOR color, FLOAT u1, FLOAT v1, FLOAT u2, FLOAT v2);
HRESULT DrawBatchRectCenter2D(FLOAT x, FLOAT y, FLOAT w, FLOAT h, D3DCOLOR color, FLOAT u1, FLOAT v1, FLOAT u2, FLOAT v2);
HRESULT PresentBatchDraw2D();

HRESULT DrawBatchLine2D(D3DXVECTOR2 &p1, D3DXVECTOR2 &p2, D3DCOLOR color1, D3DCOLOR Color2);
HRESULT PresentBatchLine2D();

//Mesh Drawr
HRESULT LoadMeshFromFile(DWORD dwNum, LPCSTR szName);
HRESULT UnloadMesh(DWORD dwNum);
HRESULT DrawMesh(D3DXMATRIX &m);
HRESULT DrawMesh(D3DXVECTOR3 &pos, D3DXVECTOR3 &rotate, D3DXVECTOR3 &scale);
HRESULT PresentMesh(DWORD dwNum);
HRESULT DrawUpMesh(DWORD num);

//Effect API
HRESULT DrawEffectGaussianBlur(D3DCOLOR color);

//Shader API
HRESULT CreateShaderEffectFromFile(LPCSTR name);
HRESULT CreateShaderEffectFromCode(LPCSTR code, DWORD len);
HRESULT AssembleShaderFromFile(LPCSTR vs_filename, LPCSTR ps_filename);
HRESULT AssembleShaderFromCode(LPCSTR vs_code, DWORD vs_len, LPCSTR ps_code, DWORD ps_len);
HRESULT CompileShaderFromFile(LPCSTR filename, LPSTR vs_name, LPSTR ps_name);
HRESULT CompileShaderFromCode(LPCSTR code, DWORD len, LPSTR vs_name, LPSTR ps_name);
HRESULT ReleaseShader();

//Set Vertex Shader
HRESULT SetShaderVertex();

//Set Pixel Shader
HRESULT SetShaderPixel();

//Set Constant
HRESULT SetShaderVertexConstantF(DWORD start, const FLOAT *data, DWORD size);
HRESULT SetShaderPixelConstantF(DWORD start, const FLOAT *data, DWORD size);

//Shader Effect
HRESULT SetShaderEffectMatrix(LPCSTR name, D3DXMATRIX &m);
HRESULT SetShaderEffectValue(LPCSTR name, LPCVOID data, DWORD len);
HRESULT SetTechniqueShaderEffect(LPCSTR name);
HRESULT BeginShaderEffect(UINT *lpPass);
HRESULT BeginShaderEffectPass(UINT Pass);
HRESULT EndShaderEffect();
HRESULT EndShaderEffectPass();

//------------------------------------------------------------------------------
// DirectX Input API
//------------------------------------------------------------------------------
DWORD GetKeyState();
HRESULT UpdateKeyState();
DWORD GetKeyboardState(int key);
DWORD GetJoystickButton(int btn);
DWORD GetJoystickUp();
DWORD GetJoystickDown();
DWORD GetJoystickLeft();
DWORD GetJoystickRight();

inline DWORD GetKeyButton0() {
  return GetKeyState() & Button_0;
}

inline DWORD GetKeyButton1() {
  return GetKeyState() & Button_1;
}


inline DWORD GetKeyButton2() {
  return GetKeyState() & Button_2;
}

inline DWORD GetKeyButton3() {
  return GetKeyState() & Button_3;
}


inline DWORD GetKeyUp() {
  return GetKeyState() & Up;
}

inline DWORD GetKeyDown() {
  return GetKeyState() & Down;
}

inline DWORD GetKeyLeft() {
  return GetKeyState() & Left;
}

inline DWORD GetKeyRight() {
  return GetKeyState() & Right;
}

//------------------------------------------------------------------------------
// DirectX Sound API
//------------------------------------------------------------------------------
inline LONG    GetSoundParam(DWORD pan, DWORD volume) {
  return ( ( (pan & 0x0000FFFF) << 16) | (volume & 0x0000FFFF) );
}

HRESULT   LoadSoundFromFile(DWORD num, LPCSTR szName);
HRESULT   LoadStreamSoundFromFile(DWORD num, LPCSTR szName);
HRESULT   UnloadSound(DWORD num);
HRESULT   SetLoopSound(DWORD num, BOOL b);
HRESULT   PlaySound(DWORD num);
HRESULT   StopSound(DWORD num);

#endif

