#include <stdio.h>
#include <stdlib.h>

unsigned long crc_table[256];
unsigned long init_crc32(unsigned char *buf, int len);
unsigned long (*s_func_crc32)(unsigned char *buf, int len) = init_crc32;

static void make_crc_table(void) {
  unsigned long c;
  int n, k;
  for (n = 0; n < 256; n++) {
    c = (unsigned long) n;
    for (k = 0; k < 8; k++) {
      if (c & 1) {
        c = 0xedb88320L ^ (c >> 1);
      } else {
        c = c >> 1;
      }
    }
    crc_table[n] = c;
  }
}

static unsigned long update_crc(unsigned long crc, unsigned char *buf, int len) {
  unsigned long c = crc ^ 0xffffffffL;
  int n;
  for (n = 0; n < len; n++) {
    c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);
  }
  return c ^ 0xffffffffL;
}

/* Return the CRC of the bytes buf[0..len-1]. */
static unsigned long static_func_crc(unsigned char *buf, int len) {
  return update_crc(0L, buf, len);
}

unsigned long init_crc32(unsigned char *buf, int len) {
  make_crc_table();
  s_func_crc32 = static_func_crc;
  return s_func_crc32(buf, len);
}

unsigned long calc_crc32(unsigned char *buf, int len) {
  return s_func_crc32(buf, len);
}
