#include "common.h"

static LPDIRECTINPUT8          lpDI          = NULL;
static LPDIRECTINPUTDEVICE8    lpInput[Max]  = { NULL };
static DIDEVICEINSTANCE        didi[Max];
static BYTE                    diks[256];
static DIJOYSTATE              dijoy;
static LONG                    RangeMax  = 1000;
static LONG                    BufferMax = 4000;

HRESULT TermDirectInput() {
  for(int i = 0 ; i < Max; i++) {
    if(lpInput[i]) {
      lpInput[i]->Unacquire();
    }
    RELEASE(lpInput[i]);
  }
  
  ZeroMemory(diks, sizeof(diks));
  ZeroMemory(didi, sizeof(didi));
  ZeroMemory(&dijoy, sizeof(dijoy));
  
  RELEASE(lpDI);
  return S_OK;
}


static HRESULT CreateKeyboard() {
  HRESULT hRet = E_FAIL;
  
  //Create Device Keyboard
  if( FAILED(hRet = lpDI->CreateDevice( GUID_SysKeyboard, &lpInput[Keyboard], NULL)) ) {
    ShowErrorMessage("CreateKeyboard : CreateDevice\n", "Error");
    return hRet;
  }

  //Set format
  if( FAILED(hRet = lpInput[Keyboard]->SetDataFormat( &c_dfDIKeyboard )) ) {
    ShowErrorMessage("CreateKeyboard : SetDataFormat\n", "Error");
    return hRet;
  }

  //xݒ
  if( FAILED( hRet = lpInput[Keyboard]->SetCooperativeLevel(GetWindowHandle(), CoopLevel)) ) {
    ShowErrorMessage("CreateKeyboard : SetCooperativeLevel\n", "Error");
    return hRet;
  }

  //foCX擾
  didi[Keyboard].dwSize = sizeof(DIDEVICEINSTANCE );
  if( FAILED( hRet = lpInput[Keyboard]->GetDeviceInfo(&didi[Keyboard]) ) ) {
    ShowErrorMessage("CreateKeyboard : GetDeviceInfo\n", "Error");
    return hRet;
  }
  
  lpInput[Keyboard]->Acquire();
  return hRet;
}

static BOOL CALLBACK DIEnumJoystickCallback(const DIDEVICEINSTANCE* lpddi, LPVOID lp) {
  HRESULT hRet;
  hRet = lpDI->CreateDevice(lpddi->guidInstance, &lpInput[Joystick], NULL);
  if FAILED(hRet) return DIENUM_CONTINUE;
  return DIENUM_STOP;
}

static HRESULT CreateJoystick() {
  HRESULT hRet = E_FAIL;
  if( FAILED(hRet = lpDI->EnumDevices(DI8DEVCLASS_GAMECTRL, DIEnumJoystickCallback, NULL, DIEDFL_ATTACHEDONLY ) ) ) {
    ShowErrorMessage("CreateJoystick : EnumDevices\n", "Error");
    return hRet;
  }
  
  if( lpInput[Joystick] == NULL ) {
    hRet = E_FAIL;
    return hRet;
  }

  if( FAILED(hRet = lpInput[Joystick]->SetDataFormat(&c_dfDIJoystick)) ) {
    ShowErrorMessage("CreateJoystick : SetDataFormat\n", "Error");
    return hRet;
  }
  
  if( FAILED( hRet = lpInput[Joystick]->SetCooperativeLevel(GetWindowHandle(), CoopLevel) ) ) {
    ShowErrorMessage("CreateJoystick : SetCooperativeLevel\n", "Error");
    return hRet;
  }

  DIPROPRANGE di;
  di.diph.dwSize = sizeof(DIPROPRANGE);
  di.diph.dwHeaderSize = sizeof(di.diph);
  

  //X
  di.diph.dwObj = DIJOFS_X; 
  di.diph.dwHow = DIPH_BYOFFSET; 
  di.lMin = -RangeMax; 
  di.lMax = +RangeMax;
  if( FAILED(hRet = lpInput[Joystick]->SetProperty( DIPROP_RANGE, &di.diph ) ) ) {
    ShowErrorMessage("CreateJoystick : SetProperty DIPROP_RANGE DIJOFS_X\n", "Error");
    return hRet;
  }

  //Y
  di.diph.dwObj = DIJOFS_Y; 
  di.diph.dwHow = DIPH_BYOFFSET; 
  di.lMin = -RangeMax; 
  di.lMax = +RangeMax;
  if( FAILED(hRet = lpInput[Joystick]->SetProperty( DIPROP_RANGE, &di.diph ) ) ) {
    ShowErrorMessage("CreateJoystick : SetProperty DIPROP_RANGE DIJOFS_Y\n", "Error");
    return hRet;
  }

  // ][lݒ
  DIPROPDWORD        dipdw;
  dipdw.diph.dwSize       = sizeof( DIPROPDWORD );
  dipdw.diph.dwHeaderSize = sizeof( dipdw.diph );
  dipdw.diph.dwHow        = DIPH_BYOFFSET;
  dipdw.dwData            = BufferMax;

  //X̖][ݒ
  dipdw.diph.dwObj    = DIJOFS_X;
  if( FAILED(hRet = lpInput[Joystick]->SetProperty(DIPROP_DEADZONE, &dipdw.diph )) ) {
    ShowErrorMessage("CreateJoystick : SetProperty DIJOFS_X\n", "Error");
    return hRet;
  }

  //Y̖][ݒ
  dipdw.diph.dwObj    = DIJOFS_Y;
  if( FAILED(hRet = lpInput[Joystick]->SetProperty(DIPROP_DEADZONE, &dipdw.diph )) ) {
    ShowErrorMessage("CreateJoystick : SetProperty DIJOFS_Y\n", "Error");
    return hRet;
  }

  //foCX擾
  didi[Joystick].dwSize = sizeof(DIDEVICEINSTANCE );
  lpInput[Joystick]->GetDeviceInfo(&didi[Joystick]);
  lpInput[Joystick]->Acquire();
  
  return hRet;
}

HRESULT InitDirectInput() {
  HRESULT hRet = E_FAIL;
  hRet = DirectInput8Create(GetInstanceHandle(), DIRECTINPUT_VERSION, IID_IDirectInput8, (LPVOID*)&lpDI, NULL);
  if(hRet != DI_OK) {
    ShowErrorMessage("InitInput : DirectInput8Create\n", "Error");
    Term();
    return hRet;
  }
  
  hRet = CreateKeyboard();
  if(hRet != DI_OK) {
    ShowErrorMessage("InitInput : CreateKeyboard\n", "Error");
    return hRet;
  }
  
  hRet = CreateJoystick();
  if(hRet != DI_OK) {
    ShowErrorMessage("InitInput : CreateJoystick\n", "Error");
    return hRet;
  }

  return hRet;
}

HRESULT UpdateKeyState() {
  HRESULT hRet = E_FAIL;
  DWORD dwState = 0;

  //Keyboard
  if(lpInput[Keyboard]) {
    ZeroMemory(diks, sizeof(diks) );
    hRet = lpInput[Keyboard]->GetDeviceState(sizeof(diks), diks);
    if( hRet != DI_OK) {
      hRet = lpInput[Keyboard]->Acquire();
    }
  }
  
  //JoyStick
  if(lpInput[Joystick]) {
    hRet = lpInput[Joystick]->Poll();
    if(hRet == DI_OK || hRet == DI_NOEFFECT) {
      hRet = lpInput[Joystick]->GetDeviceState( sizeof(DIJOYSTATE), &dijoy );
    }
    if( FAILED( hRet ) ) {
      lpInput[Joystick]->Acquire();
    }
  }
  
  return hRet;
}

DWORD ConvertJoystickState() {
  DWORD      dwState = 0;
  if(dijoy.lY < 0)                dwState |= Up;
  if(dijoy.lY > 0)                dwState |= Down;
  if(dijoy.lX < 0)                dwState |= Left;
  if(dijoy.lX > 0)                dwState |= Right;
  if(dijoy.rgbButtons[0] & 0x80)  dwState |= Button_0;
  if(dijoy.rgbButtons[1] & 0x80)  dwState |= Button_1;
  if(dijoy.rgbButtons[2] & 0x80)  dwState |= Button_2;
  if(dijoy.rgbButtons[3] & 0x80)  dwState |= Button_3;
  if(dijoy.rgbButtons[4] & 0x80)  dwState |= Button_4;
  if(dijoy.rgbButtons[5] & 0x80)  dwState |= Button_5;
  if(dijoy.rgbButtons[6] & 0x80)  dwState |= Button_6;
  if(dijoy.rgbButtons[7] & 0x80)  dwState |= Button_7;
  if(dijoy.rgbButtons[8] & 0x80)  dwState |= Button_8;
  return dwState;
}



DWORD ConvertKeyboardState() {
  DWORD      dwState = 0;
  if(diks[DIK_UP]     == 0x80)  dwState |= Up;
  if(diks[DIK_DOWN]   == 0x80)  dwState |= Down;
  if(diks[DIK_LEFT]   == 0x80)  dwState |= Left;
  if(diks[DIK_RIGHT]  == 0x80)  dwState |= Right;
  if(diks[DIK_Z]      == 0x80)  dwState |= Button_0;
  if(diks[DIK_X]      == 0x80)  dwState |= Button_1;
  if(diks[DIK_C]      == 0x80)  dwState |= Button_2;
  if(diks[DIK_A]      == 0x80)  dwState |= Button_3;
  if(diks[DIK_S]      == 0x80)  dwState |= Button_4;
  if(diks[DIK_D]      == 0x80)  dwState |= Button_5;
  if(diks[DIK_Q]      == 0x80)  dwState |= Button_6;
  if(diks[DIK_W]      == 0x80)  dwState |= Button_7;
  if(diks[DIK_E]      == 0x80)  dwState |= Button_8;

  return dwState;
}

DWORD GetKeyboardState(int key) {
  if(key < 256 && key > 0) {
    if(diks[key]     == 0x80)  return 1;
  }
  return 0;
}

DWORD GetJoystickButton(int btn) {
  if(dijoy.rgbButtons[btn] & 0x80) return 1;
  return 0;
}

DWORD GetJoystickUp() {
  if(dijoy.lY < 0) return 1;
  return 0;
}

DWORD GetJoystickDown() {
  if(dijoy.lY > 0) return 1;
  return 0;
}

DWORD GetJoystickLeft() {
  if(dijoy.lX < 0) return 1;
  return 0;
}

DWORD GetJoystickRight() {
  if(dijoy.lX > 0) return 1;
  return 0;
}

DWORD GetKeyState() {
  return ConvertKeyboardState() | ConvertJoystickState();
}
