//------------------------------------------------------------------------------
//
// main.cpp
//
//------------------------------------------------------------------------------
#include "common.h"
#include <lua/lua.hpp>

static lua_State* L = NULL;

FLOAT GetFloatRand() {
  return ((FLOAT)rand()) / 32768.0f;
}

void DispVector3(Vector3 &v) {
  printf("%08X : %f %f %f\n", &v, v.x, v.y, v.z);
}

//------------------------------------------------------------------------------
//
// Lua function
//
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// SYSTEM
//------------------------------------------------------------------------------
int l_test(lua_State* L) {
  FLOAT a = (FLOAT)luaL_checknumber(L, 1);
  FLOAT b = (FLOAT)luaL_checknumber(L, 2);
  FLOAT c = (FLOAT)luaL_checknumber(L, 3);
  
  printf("Lua Test : %f %f %f\n", a, b, c);
  return 1;
}

int l_init(lua_State* L) {
  const char *title   = luaL_checkstring (L, 1);
  DWORD Width         = luaL_checkint(L, 2);
  DWORD Height        = luaL_checkint(L, 3);
  BOOL  bIsWindow     = luaL_checkint(L, 4);
  HRESULT hRet        = Init(title, Width, Height, bIsWindow);
  lua_pushnumber(L, hRet);
  return 1;
}

int l_term(lua_State* L) {
  Term();
  return 0;
}

int l_do_event(lua_State* L) {
  lua_pushnumber(L, DoEvent());
  return 1;
}


int l_start_scene(lua_State* L) {
  D3DCOLOR temp = (unsigned long)luaL_checklong(L, 1);
  //D3DCOLOR temp = (unsigned long)lua_touserdata(L, -1);
  D3DCOLOR col = (D3DCOLOR)temp | 0xFF000000;
  lua_pushnumber(L, StartScene(col, FALSE) );
  return 1;
}

int l_end_scene(lua_State* L) {
  EndScene();
  return 0;
}

int l_flip(lua_State* L) {
  Flip();
  return 0;
}

//------------------------------------------------------------------------------
// RENDER
//------------------------------------------------------------------------------
int l_load_texture(lua_State* L) {
  const char *name   = luaL_checkstring (L, 1);
  HRESULT hRet = LoadTextureFromFile(name);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_unload_texture(lua_State* L) {
  const char *name   = luaL_checkstring (L, 1);
  HRESULT hRet = UnloadTexture(name);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_set_texture(lua_State* L) {
  const char *name   = luaL_checkstring (L, 1);
  HRESULT hRet = SetTexture(name);
  lua_pushnumber(L, hRet );
  return 1;
}

//UV
static FLOAT g_u1 = 0.0f;
static FLOAT g_v1 = 0.0f;
static FLOAT g_u2 = 1.0f;
static FLOAT g_v2 = 1.0f;

int l_set_uv_rect2d(lua_State* L) {
  g_u1 = (FLOAT)luaL_checknumber(L, 1);
  g_v1 = (FLOAT)luaL_checknumber(L, 2);
  g_u2 = (FLOAT)luaL_checknumber(L, 3);
  g_v2 = (FLOAT)luaL_checknumber(L, 4);
  return 0;
}

int l_draw_rect2d_up(lua_State* L) {
  LONG temp = luaL_checklong(L, 5);
  D3DCOLOR col = (D3DCOLOR)temp;
  LONG temp2 = Sat8Bit(luaL_checklong(L, 6));

  col |= (temp2 << 24);
  DrawRect2D(
    (FLOAT)luaL_checknumber(L, 1), //x
    (FLOAT)luaL_checknumber(L, 2), //y
    (FLOAT)luaL_checknumber(L, 3), //w
    (FLOAT)luaL_checknumber(L, 4), //h
    col, g_u1, g_v1, g_u2, g_v2);
  return 0;
}


int l_draw_rect2d(lua_State* L) {
  LONG temp = luaL_checklong(L, 5);
  D3DCOLOR col = (D3DCOLOR)temp;
  LONG temp2 = Sat8Bit(luaL_checklong(L, 6));
  col |= (temp2 << 24);
  DrawBatchRect2D(
    (FLOAT)luaL_checknumber(L, 1), //x
    (FLOAT)luaL_checknumber(L, 2), //y
    (FLOAT)luaL_checknumber(L, 3), //w
    (FLOAT)luaL_checknumber(L, 4), //h
    col, g_u1, g_v1, g_u2, g_v2);
  return 0;
}

int l_draw_center_rect_2d(lua_State* L) {
  LONG temp = luaL_checklong(L, 5);
  D3DCOLOR col = (D3DCOLOR)temp;
  LONG temp2 = Sat8Bit(luaL_checklong(L, 6));
  col |= (temp2 << 24);
  DrawBatchRectCenter2D(
    (FLOAT)luaL_checknumber(L, 1), //x
    (FLOAT)luaL_checknumber(L, 2), //y
    (FLOAT)luaL_checknumber(L, 3), //w
    (FLOAT)luaL_checknumber(L, 4), //h
    col, g_u1, g_v1, g_u2, g_v2);
  return 0;
}

int l_present_rect_2d(lua_State* L) {
  PresentBatchDraw2D();
  return 0;
}

int l_load_mesh(lua_State* L) {
  LONG temp = luaL_checklong(L, 1);
  const char *name   = luaL_checkstring (L, 2);
  HRESULT hRet = LoadMeshFromFile(temp, name);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_unload_mesh(lua_State* L) {
  LONG temp = luaL_checklong(L, 1);
  HRESULT hRet = UnloadMesh(temp);
  lua_pushnumber(L, hRet );
  return 1;
}

//
// px, px, px, 
// rx, rx, rx
// sx, sx, sx
int l_draw_mesh(lua_State* L) {
  D3DXVECTOR3 pos(
    (FLOAT)luaL_checknumber(L, 1), 
    (FLOAT)luaL_checknumber(L, 2), 
    (FLOAT)luaL_checknumber(L, 3));
  
  D3DXVECTOR3 rotate(
    (FLOAT)luaL_checknumber(L, 4), 
    (FLOAT)luaL_checknumber(L, 5), 
    (FLOAT)luaL_checknumber(L, 6));
  
  D3DXVECTOR3 scale(
    (FLOAT)luaL_checknumber(L, 7), 
    (FLOAT)luaL_checknumber(L, 8), 
    (FLOAT)luaL_checknumber(L, 9));

  HRESULT hRet = DrawMesh(pos, rotate, scale);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_present_mesh(lua_State* L) {
  LONG temp = luaL_checklong(L, 1);
  HRESULT hRet = PresentMesh(temp);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_set_projection(lua_State* L) {
  FLOAT ffov     = (FLOAT)luaL_checknumber(L, 1);
  FLOAT faspect  = (FLOAT)luaL_checknumber(L, 2);
  FLOAT fnear    = (FLOAT)luaL_checknumber(L, 3);
  FLOAT ffar     = (FLOAT)luaL_checknumber(L, 4);
  HRESULT hRet = SetProjection(ffov, faspect, fnear, ffar);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_set_view(lua_State* L) {
  D3DXVECTOR3 pos(
    (FLOAT)luaL_checknumber(L, 1), 
    (FLOAT)luaL_checknumber(L, 2), 
    (FLOAT)luaL_checknumber(L, 3));
  
  D3DXVECTOR3 at(
    (FLOAT)luaL_checknumber(L, 4), 
    (FLOAT)luaL_checknumber(L, 5), 
    (FLOAT)luaL_checknumber(L, 6));
  
  D3DXVECTOR3 up(
    (FLOAT)luaL_checknumber(L, 7), 
    (FLOAT)luaL_checknumber(L, 8), 
    (FLOAT)luaL_checknumber(L, 9));
  HRESULT hRet = SetView(pos, at, up);
  lua_pushnumber(L, hRet );
  return 1;
}


int l_set_light(lua_State* L) {
  BOOL     bEnable = luaL_checklong(L, 1);
  D3DCOLOR ambient = luaL_checklong(L, 2);
  D3DXCOLOR col(
    (FLOAT)luaL_checknumber(L, 3), 
    (FLOAT)luaL_checknumber(L, 4), 
    (FLOAT)luaL_checknumber(L, 5),
    (FLOAT)luaL_checknumber(L, 6)
  );
  
  D3DXVECTOR3 direction(
    (FLOAT)luaL_checknumber(L, 7), 
    (FLOAT)luaL_checknumber(L, 8), 
    (FLOAT)luaL_checknumber(L, 9)
  );
  
  HRESULT hRet = SetLight(bEnable, ambient, col, direction);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_set_ambient(lua_State* L) {
  D3DCOLOR ambient = luaL_checklong(L, 1);
  HRESULT hRet = SetAmbient(ambient);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_set_render_state(lua_State *L) {
  LONG temp = luaL_checklong(L, 1);
  HRESULT hRet = SetRenderState(temp);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_set_fog_param(lua_State *L) {
  HRESULT hRet = SetFogParam((FLOAT)luaL_checknumber(L, 1), (FLOAT)luaL_checknumber(L, 2), (D3DCOLOR)luaL_checklong(L, 3) );
  lua_pushnumber(L, hRet );
  return 1;
}

//font
int l_create_fontdata(lua_State *L) {
  DWORD temp         = luaL_checklong(L, 1);
  const char *name   = luaL_checkstring (L, 2);
  UINT        pitch  = luaL_checklong (L, 3);
  HRESULT hRet       = CreateFontData(temp, name, pitch);
  lua_pushnumber(L, hRet );
  return 1;
}

int l_release_fontdata(lua_State *L) {
  DWORD temp         = luaL_checklong(L, 1);
  HRESULT hRet       = ReleaseFontData(temp);
  lua_pushnumber(L, hRet );
  return 1;
}

//ϒǂȂ
int l_draw_fontdata(lua_State *L) {
  DWORD    temp           = luaL_checklong(L, 1);
  FLOAT    x              = (FLOAT)luaL_checknumber(L, 2);
  FLOAT    y              = (FLOAT)luaL_checknumber(L, 3);
  D3DCOLOR color          = luaL_checklong(L, 4);
  D3DCOLOR alpha          = luaL_checklong(L, 5);
  const char *name        = luaL_checkstring (L, 6);
  HRESULT hRet = DrawFontData(temp, x, y, color | (alpha << 24), 1, name);
  lua_pushnumber(L, hRet );
  return 1;
}

//------------------------------------------------------------------------------
// SOUND
//------------------------------------------------------------------------------
int l_sound_load_from_file(lua_State* L) {
  LONG num = luaL_checklong(L, 1);
  const char *name   = luaL_checkstring (L, 2);
  HRESULT result = LoadSoundFromFile(num, name);
  lua_pushnumber(L, result );
  return 1;
}

int l_sound_load_stream_from_file(lua_State* L) {
  LONG num = luaL_checklong(L, 1);
  const char *name   = luaL_checkstring (L, 2);
  HRESULT result = LoadStreamSoundFromFile(num, name);
  lua_pushnumber(L, result );
  return 1;
}

int l_sound_unload(lua_State* L) {
  LONG num = luaL_checklong(L, 1);
  HRESULT result = UnloadSound(num);
  lua_pushnumber(L, result );
  return 1;
}


int l_sound_play(lua_State* L) {
  LONG num = luaL_checklong(L, 1);
  HRESULT result = PlaySound(num);
  lua_pushnumber(L, result );
  return 1;
}

int l_sound_stop(lua_State* L) {
  LONG num = luaL_checklong(L, 1);
  HRESULT result = StopSound(num);
  lua_pushnumber(L, result );
  return 1;
}


//------------------------------------------------------------------------------
// INPUT
//------------------------------------------------------------------------------
int l_input_button0(lua_State* L) {
  lua_pushnumber(L, GetKeyButton0() );
  return 1;
}

int l_input_button1(lua_State* L) {
  lua_pushnumber(L, GetKeyButton1() );
  return 1;
}

int l_input_button2(lua_State* L) {
  lua_pushnumber(L, GetKeyButton2() );
  return 1;
}

int l_input_button3(lua_State* L) {
  lua_pushnumber(L, GetKeyButton3() );
  return 1;
}

int l_input_up(lua_State* L) {
  lua_pushnumber(L, GetKeyUp() );
  return 1;
}

int l_input_down(lua_State* L) {
  lua_pushnumber(L, GetKeyDown() );
  return 1;
}

int l_input_left(lua_State* L) {
  lua_pushnumber(L, GetKeyLeft() );
  return 1;
}

int l_input_right(lua_State* L) {
  lua_pushnumber(L, GetKeyRight() );
  return 1;
}


//------------------------------------------------------------------------------
// Function Table
//------------------------------------------------------------------------------
//system
static luaL_Reg lua_func_table[] = {
  {"test",                    l_test},
  {"init",                    l_init},
  {"term",                    l_term},
  {"do_event",                l_do_event},
  {"start_scene",             l_start_scene},
  {"end_scene",               l_end_scene},
  {"flip",                    l_flip},
  {NULL,                      NULL},
};

//render
static luaL_Reg lua_func_draw_table[] = {
  //texture
  {"load_texture",            l_load_texture},
  {"unload_texture",          l_unload_texture},
  {"set_texture",             l_set_texture},
  {"set_texture",             l_set_texture},
  
  //draw 2d rect
  {"set_uv_rect2d",           l_set_uv_rect2d},
  {"rect2d_up",               l_draw_rect2d_up},
  {"rect2d",                  l_draw_rect2d},
  {"center_rect_2d",          l_draw_center_rect_2d},
  {"present_rect_2d",         l_present_rect_2d},
  
  //mesh
  {"load_mesh",               l_load_mesh},
  {"unload_mesh",             l_unload_mesh},
  {"draw_mesh",               l_draw_mesh},
  {"present_mesh",            l_present_mesh},
  
  //matrix
  {"set_projection",          l_set_projection},
  {"set_view",                l_set_view},
  
  //light fog misc
  {"set_light",                l_set_light},
  {"set_render_state",         l_set_render_state},
  {"set_fog_param",            l_set_fog_param},
  
  //font
  {"create_fontdata",           l_create_fontdata},
  {"release_fontdata",          l_release_fontdata},
  {"draw_fontdata",             l_draw_fontdata},
  
  {NULL,                      NULL},
};

//sound
static luaL_Reg lua_func_sound_table[] = {
  {"load_from_file",          l_sound_load_from_file},
  {"load_stream_from_file",   l_sound_load_stream_from_file},
  {"unload",                  l_sound_unload},
  {"play",                    l_sound_play},
  {"stop",                    l_sound_stop},
  {NULL,                      NULL},
};

//input
static luaL_Reg lua_func_input_table[] = {
  {"button0",                 l_input_button0},
  {"button1",                 l_input_button1},
  {"button2",                 l_input_button2},
  {"button3",                 l_input_button3},
  {"up",                      l_input_up},
  {"down",                    l_input_down},
  {"left",                    l_input_left},
  {"right",                   l_input_right},
  {NULL,                      NULL},
};



int main(int argc, char *argv[]) {
  //Initialize Lua
  L = luaL_newstate();
  if(L) {
    luaL_openlibs(L);
    
    //Register table
    luaL_register(L, "system", lua_func_table);
    luaL_register(L, "render", lua_func_draw_table);
    luaL_register(L, "sound",  lua_func_sound_table);
    luaL_register(L, "input",  lua_func_input_table);
    if(argc < 2) {
      if(luaL_dofile(L, "boot.lua")) {
        if(luaL_dofile(L, "boot.out")) {
          if(luaL_dofile(L, "boot.bin")) {
            printf("Can't do boot file\n");
          }
        }
      }
    } else {
      if(luaL_dofile(L, argv[1])) {
        printf("Can't do %s file\n", argv[1]);
      }
    }
    
    //Terminate Lua
    lua_close(L);
  } else {
    printf("FAILED.\n");
  }
  Term();
  return 0;
}
