//------------------------------------------------------------------------------
//
// shader.cpp
//
//------------------------------------------------------------------------------
#include "common.h"

//Shader
/*
#define VertexShader                                LPDIRECT3DVERTEXSHADER9
#define PixelShader                                 LPDIRECT3DPIXELSHADER9
#define ShaderBuffer                                LPD3DXBUFFER
*/

class Shader {
  LPD3DXEFFECT              lpEffect;
  LPDIRECT3DVERTEXSHADER9   lpVertexShader;
  LPD3DXBUFFER              lpVertexShaderBuffer;
  LPDIRECT3DPIXELSHADER9    lpPixelShader;
  LPD3DXBUFFER              lpPixelShaderBuffer;
  
  HRESULT CreateShader() {
    HRESULT hRet = S_OK;
    
    if(lpVertexShaderBuffer) {
      if( FAILED ( hRet = GetDevice()->CreateVertexShader((const DWORD*)lpVertexShaderBuffer->GetBufferPointer(), &lpVertexShader) ) ) {
        ShowErrorMessage("Shader::CreateShader : CreateVertexShader \n", "Error");
        Release();
        return hRet;
      }
    }
    
    if(lpPixelShaderBuffer) {
      if( FAILED ( hRet = GetDevice()->CreatePixelShader((const DWORD*)lpPixelShaderBuffer->GetBufferPointer(), &lpPixelShader) ) ) {
        ShowErrorMessage("Shader::CreateShader : CreatePixelShader\n", "Error");
        Release();
        return hRet;
      }
    }
    
    /*
    //foCXXg΍̂߂ɍ폜
    if( FAILED(ReleaseBuffer()) ) {
      return E_FAIL;
    }
    */
    
    return hRet;
  }
  
  HRESULT ReleaseBuffer() {
    RELEASE(lpPixelShaderBuffer);
    RELEASE(lpVertexShaderBuffer);
    return S_OK;
  }
  
public:
  Shader() {
    lpEffect              = NULL;
    lpVertexShader        = NULL;
    lpPixelShader         = NULL;
    lpVertexShaderBuffer  = NULL;
    lpPixelShaderBuffer   = NULL;
  }
  
  ~Shader() {
    Release();
  }
  
  HRESULT DispHandle() {
    printf("eft : %08X\n", lpEffect);
    printf("vsh : %08X\n", lpVertexShader);
    printf("psh : %08X\n", lpPixelShader);
    return S_OK;
  }
  
  HRESULT ReleaseEffect() {
    RELEASE(lpEffect);
    return S_OK;
  }
  
  HRESULT ReleaseShader() {
    RELEASE(lpPixelShader);
    RELEASE(lpVertexShader);
    return S_OK;
  }
  
  HRESULT OnLost() {
    HRESULT hRet = S_OK;
    
    if( FAILED(hRet = ReleaseShader()) ) {
      return hRet;
    }
    
    if(lpEffect) {
      if( FAILED(hRet = lpEffect->OnLostDevice()) ) {
        return hRet;
      }
    }
    return hRet;
  }
  
  HRESULT OnRestore() {
    HRESULT hRet = S_OK;
    
    if( FAILED(hRet = CreateShader())) {
      return hRet;
    }
    
    if(lpEffect) {
      if( FAILED(hRet = lpEffect->OnResetDevice()) ) {
        return hRet;
      }
    }
    return hRet;
  }
  
  HRESULT Release() {
    return ReleaseBuffer() | ReleaseShader() | ReleaseEffect();
  }
  
  //Set Matrix
  HRESULT SetMatrixEffect(LPCSTR name, D3DXMATRIX &m) {
    if(!lpEffect) return E_FAIL;
    return lpEffect->SetMatrix(name, &m);
  }
  
  //Set Value
  HRESULT SetValueEffect(LPCSTR name, LPCVOID data, DWORD len) {
    if(!lpEffect) return E_FAIL;
    return lpEffect->SetValue(name, data, len);
  }
  
  //Set Technique
  HRESULT SetEffectTechnique(LPCSTR name) {
    if(!lpEffect) return E_FAIL;
    return lpEffect->SetTechnique(name);
  }
  
  //Begin Effect
  HRESULT BeginEffect(UINT *lpPass) {
    if(!lpEffect) return E_FAIL;
    return lpEffect->Begin(lpPass, 0);
  }
  
  //Begin Pass
  HRESULT BeginEffectPass(UINT Pass) {
    if(!lpEffect) return E_FAIL;
    return lpEffect->BeginPass(Pass);
  }
  
  //End Pass
  HRESULT EndEffect() {
    if(!lpEffect) return E_FAIL;
    return lpEffect->End();
  }
  
  //End Pass
  HRESULT EndEffectPass() {
    if(!lpEffect) return E_FAIL;
    return lpEffect->EndPass();
  }
  
  //Set Vertex Shader
  HRESULT SetVertexShader() {
    if(GetDevice() == NULL || !lpVertexShader) return E_FAIL;
    return GetDevice()->SetVertexShader(lpVertexShader);
  }
  
  //Set Vertex ConstantF
  HRESULT SetVertexConstantF(DWORD start, const FLOAT *lpdata, DWORD size) {
    if(GetDevice() == NULL || !lpVertexShader) return E_FAIL;
    return GetDevice()->SetVertexShaderConstantF(start, lpdata, size);
  }
  
  //Set Pixel Shader
  HRESULT SetPixelShader() {
    if(GetDevice() == NULL || !lpPixelShader) return E_FAIL;
    return GetDevice()->SetPixelShader(lpPixelShader);
  }
  
  //Set Pixel ConstantF
  HRESULT SetPixelConstantF(DWORD start, const FLOAT *lpdata, DWORD size) {
    if(GetDevice() == NULL || !lpPixelShader) return E_FAIL;
    return GetDevice()->SetPixelShaderConstantF(start, lpdata, size);
  }
  
  //Create Effect From File
  HRESULT CreateEffectFromFile(LPCSTR filename) {
    HRESULT hRet = E_FAIL;
    if(GetDevice() == NULL || !filename ) return E_FAIL;
    
    ReleaseEffect();
    if( FAILED ( hRet = D3DXCreateEffectFromFile(GetDevice(), filename, 0, 0, 0, 0, &lpEffect, 0) ) ) {
      ShowErrorMessage("Shader::CreateEffectFromFile : D3DXCreateEffectFromFile\n", "Error");
      ReleaseEffect();
      return hRet;
    }
    
    return hRet;
  }
  
  //Create Effect From Code
  HRESULT CreateEffectFromCode(LPCSTR code, DWORD len) {
    HRESULT hRet = E_FAIL;
    if(GetDevice() == NULL || !code) return E_FAIL;
    
    if(code && (!len || len < 0)) {
      len = strlen(code);
    }
    
    ReleaseEffect();
    if( FAILED ( hRet = D3DXCreateEffect(GetDevice(), code, len, 0, 0, 0, 0, &lpEffect, 0) ) ) {
      ShowErrorMessage("Shader::CreateEffectFromFile : D3DXCreateEffectFromFile\n", "Error");
      ReleaseEffect();
      return hRet;
    }
    
    return hRet;
  }
  
  HRESULT AssembleFromFile(LPCSTR vs_filename, LPCSTR ps_filename) {
    HRESULT hRet = E_FAIL;
    if(GetDevice() == NULL || (!vs_filename && !ps_filename) ) return E_FAIL;
    
    ReleaseShader();
    
    //Create VertexShader
    if(vs_filename) {
      RELEASE(lpVertexShader);
      if( FAILED ( hRet = D3DXAssembleShaderFromFile(vs_filename, 0, 0, 0, &lpVertexShaderBuffer, 0)) ) {
        ShowErrorMessage("Shader::AssembleFromFile : D3DXAssembleShader VertexShader \n", "Error");
        ReleaseShader();
        return hRet;
      }
    }
    
    //Create PixelShader
    if(ps_filename) {
      RELEASE(lpPixelShader);
      if( FAILED ( hRet = D3DXAssembleShaderFromFile(ps_filename, 0, 0, 0, &lpPixelShaderBuffer, 0)) ) {
        ShowErrorMessage("Shader::AssembleFromFile : D3DXAssembleShader PixelShader \n", "Error");
        ReleaseShader();
        return hRet;
      }
    }
    
    //Create Do.
    return CreateShader();
  }
  
  HRESULT AssembleFromCode(LPCSTR vs_code, DWORD vs_len, LPCSTR ps_code, DWORD ps_len) {
    HRESULT hRet = E_FAIL;
    
    if(GetDevice() == NULL || (!vs_code && !ps_code) ) return E_FAIL;
    
    //Release
    ReleaseShader();
    
    //If not use length, then calc code length.
    if(vs_code && (!vs_len || vs_len < 0)) {
      vs_len = strlen(vs_code);
    }
    
    //If not use length, then calc code length.
    if(ps_code && (!ps_len || ps_len < 0)) {
      ps_len = strlen(ps_code);
    }
    
    //Create VertexShader
    if(vs_code) {
      if( FAILED ( hRet = D3DXAssembleShader(vs_code, vs_len, 0, 0, 0, &lpVertexShaderBuffer, 0)) ) {
        ShowErrorMessage("Shader::AssembleFromCode : D3DXAssembleShader VertexShader \n", "Error");
        ReleaseShader();
        return hRet;
      }
    }
    
    //Create PixelShader
    if(ps_code) {
      if( FAILED ( hRet = D3DXAssembleShader(ps_code, ps_len, 0, 0, 0, &lpPixelShaderBuffer, 0)) ) {
        ShowErrorMessage("Shader::AssembleFromCode : D3DXAssembleShader PixelShader \n", "Error");
        ReleaseShader();
        return hRet;
      }
    }
    
    //Create Do.
    return CreateShader();
  }
  
  HRESULT CompileFromFile(LPCSTR filename, LPSTR vs_name, LPSTR ps_name) {
    HRESULT hRet = E_FAIL;
    
    if(GetDevice() == NULL || !filename || (!vs_name && !ps_name) ) return E_FAIL;
    
    //Release
    ReleaseShader();
    
    //Create VertexShader
    if(vs_name) {
      if( FAILED ( hRet = D3DXCompileShaderFromFile(filename, 0, 0, vs_name, D3DXGetVertexShaderProfile(GetDevice()), 0, &lpVertexShaderBuffer, 0, 0)) ) {
        ShowErrorMessage("Shader::CompileFromFile : D3DXCompileShader VertexShader \n", "Error");
        ReleaseShader();
        return hRet;
      }
    }
    
    //Create PixelShader
    if(ps_name) {
      if( FAILED ( hRet = D3DXCompileShaderFromFile(filename, 0, 0, ps_name, D3DXGetPixelShaderProfile(GetDevice()), 0, &lpPixelShaderBuffer, 0, 0)) ) {
        ShowErrorMessage("Shader::CompileFromFile : D3DXCompileShader PixelShader \n", "Error");
        ReleaseShader();
        return hRet;
      }
    }

    
    //Create Do.
    return CreateShader();
  }
  
  HRESULT CompileFromCode(LPCSTR code, DWORD len, LPSTR vs_name, LPSTR ps_name) {
    HRESULT hRet = E_FAIL;
    
    if(GetDevice() == NULL || !code || (!vs_name && !ps_name) ) return E_FAIL;
    
    //Release
    ReleaseShader();
    
    //If not use length, then calc code length.
    if(code && (!len || len < 0)) {
      len = strlen(code);
    }
    
    //Create VertexShader
    if(vs_name) {
      if( FAILED ( hRet = D3DXCompileShader(code, len, 0, 0, vs_name, D3DXGetVertexShaderProfile(GetDevice()), 0, &lpVertexShaderBuffer, 0, 0)) ) {
        ShowErrorMessage("Shader::CompileFromCode : D3DXCompileShader VertexShader \n", "Error");
        ReleaseShader();
        return hRet;
      }
    }
    
    //Create PixelShader
    if(ps_name) {
      if( FAILED ( hRet = D3DXCompileShader(code, len, 0, 0, ps_name, D3DXGetPixelShaderProfile(GetDevice()), 0, &lpPixelShaderBuffer, 0, 0)) ) {
        ShowErrorMessage("Shader::CompileFromCode : D3DXCompileShader PixelShader \n", "Error");
        ReleaseShader();
        return hRet;
      }
    }
    
    //Create Do.
    return CreateShader();
  }
};


static Shader shader[ShaderMax];
static DWORD  dwCurrentIndex = 0;

HRESULT OnLostShader() {
  return shader[dwCurrentIndex].OnLost();
}

HRESULT OnRestoreShader() {
  return shader[dwCurrentIndex].OnRestore();
}

HRESULT CreateShaderEffectFromFile(LPCSTR name) {
  return shader[dwCurrentIndex].CreateEffectFromFile(name);
}

HRESULT CreateShaderEffectFromCode(LPCSTR code, DWORD len) {
  return shader[dwCurrentIndex].CreateEffectFromCode(code, len);
}

HRESULT AssembleShaderFromFile(LPCSTR vs_filename, LPCSTR ps_filename) {
  return shader[dwCurrentIndex].AssembleFromFile(vs_filename, ps_filename);
}

HRESULT AssembleShaderFromCode(LPCSTR vs_code, DWORD vs_len, LPCSTR ps_code, DWORD ps_len) {
  return shader[dwCurrentIndex].AssembleFromCode(vs_code, vs_len, ps_code, ps_len);
}

HRESULT CompileShaderFromFile(LPCSTR filename, LPSTR vs_name, LPSTR ps_name) {
  return shader[dwCurrentIndex].CompileFromFile(filename, vs_name, ps_name);
}

HRESULT CompileShaderFromCode(LPCSTR code, DWORD len, LPSTR vs_name, LPSTR ps_name) {
  return shader[dwCurrentIndex].CompileFromCode(code, len, vs_name, ps_name);
}

HRESULT ReleaseShader() {
  return shader[dwCurrentIndex].Release();
}

HRESULT SetTechniqueShaderEffect(LPCSTR name) {
  return shader[dwCurrentIndex].SetEffectTechnique(name);
}

HRESULT BeginShaderEffect(UINT *lpPass) {
  return shader[dwCurrentIndex].BeginEffect(lpPass);
}

HRESULT BeginShaderEffectPass(UINT Pass) {
  return shader[dwCurrentIndex].BeginEffectPass(Pass);
}

HRESULT EndShaderEffect() {
  return shader[dwCurrentIndex].EndEffect();
}

HRESULT EndShaderEffectPass() {
  return shader[dwCurrentIndex].EndEffectPass();
}

HRESULT SetShaderEffectMatrix(LPCSTR name, D3DXMATRIX &m) {
  return shader[dwCurrentIndex].SetMatrixEffect(name, m);
}

HRESULT SetShaderEffectValue(LPCSTR name, LPCVOID data, DWORD len) {
  return shader[dwCurrentIndex].SetValueEffect(name, data, len);
}

//Set Vertex Shader
HRESULT SetShaderVertex() {
  return shader[dwCurrentIndex].SetVertexShader();
}

//Set Pixel Shader
HRESULT SetShaderPixel() {
  return shader[dwCurrentIndex].SetPixelShader();
}


//Set Constant
HRESULT SetShaderVertexConstantF(DWORD start, const FLOAT *data, DWORD size) {
  return shader[dwCurrentIndex].SetVertexConstantF(start, data, size);
}

//Set Constant
HRESULT SetShaderPixelConstantF(DWORD start, const FLOAT *data, DWORD size) {
  return shader[dwCurrentIndex].SetPixelConstantF(start, data, size);
}



//------------------------------------------------------------------------------
//
// Test Code
//
//------------------------------------------------------------------------------
static const char EffectShaderCode[] = 
"  float4x4 mWorldViewProj;                             "
"  float4   ColorDiffuse;                               "
"                                                       "
"  struct vs_out {                                      "
"    float4 pos : POSITION;                             "
"    float4 col : COLOR0;                               "
"  };                                                   "
"                                                       "
"  struct ps_out {                                      "
"    float4 col : COLOR0;                               "
"  };                                                   "
"                                                       "
"  vs_out basic_vs(float4 pos : POSITION) {             "
"    vs_out ret;                                        "
"                                                       "
"    ret.pos = mul(pos, mWorldViewProj);                "
"                                                       "
"    ret.col.rgb = ColorDiffuse;                        "
"    ret.col.a   = 1.0f;                                "
"                                                       "
"    return ret;                                        "
"  }                                                    "
"                                                       "
"  ps_out basic_ps(vs_out input) {                      "
"    ps_out ret = (ps_out)0;                            "
"                                                       "
"    ret.col.rgb = input.col;                           "
"                                                       "
"    return ret;                                        "
"  }                                                    "
"                                                       "
"technique test_effect                                  "
"{                                                      "
"  pass P0                                              "
"  {                                                    "
"    VertexShader = compile vs_2_0 basic_vs();          "
"    PixelShader  = compile ps_2_0 basic_ps();          "
"  }                                                    "
"}                                                      ";


static const char shaderCode[] = 
"  float4x4 mWorldViewProj;                             "
"  float4   ColorDiffuse;                               "
"                                                       "
"  struct vs_out {                                      "
"    float4 pos : POSITION;                             "
"    float4 col : COLOR0;                               "
"  };                                                   "
"                                                       "
"  struct ps_out {                                      "
"    float4 col : COLOR0;                               "
"  };                                                   "
"                                                       "
"  vs_out basic_vs(float4 pos : POSITION) {             "
"    vs_out ret;                                        "
"                                                       "
"    ret.pos = mul(pos, mWorldViewProj);                "
"                                                       "
"    ret.col.rgb = ColorDiffuse;                        "
"    ret.col.a   = 1.0f;                                "
"                                                       "
"    return ret;                                        "
"  }                                                    "
"                                                       "
"  ps_out basic_ps(vs_out input) {                      "
"    ps_out ret = (ps_out)0;                            "
"                                                       "
"    ret.col.rgb = input.col;                           "
"                                                       "
"    return ret;                                        "
"  }                                                    "
"                                                       ";

const char VShader[] =
  "vs_1_1                              \n"
  "dcl_position    v0                  \n"
  "dcl_color       v1                  \n"
  "m4x4            oPos,   v0,     c0  \n"
  "add             oD0 ,   v1,     c4  \n";

const char PxShader[] =
  "ps_1_1                                        \n"
  "def   c0, 0.2989f, 0.5866f, 0.1145f, 0.0f     \n"
  "tex   t0                                      \n"
  "dp3   r0, t0, c0                              \n"
  "mov   r1, r0.a                                \n"
  "lrp   r0, c1, t0, r1                          \n";

void TestShader() {
  HRESULT hRet = S_OK;
  
  Shader shader;

  hRet |= shader.CreateEffectFromFile("etest.fx");
  printf("CreateEffectFromFile : %x\n", hRet);
  shader.DispHandle();
  shader.Release();
  
  hRet |= shader.CreateEffectFromCode(EffectShaderCode, sizeof(EffectShaderCode));
  printf("CreateEffectFromFile : %x\n", hRet);
  shader.DispHandle();
  shader.Release();
  
  hRet |= shader.AssembleFromFile("vs_test.vsh", "ps_test.psh");
  printf("AssembleFromFile : %x\n", hRet);
  shader.DispHandle();
  shader.Release();
  
  hRet |= shader.AssembleFromCode(VShader, sizeof(VShader), PxShader, sizeof(PxShader) );
  printf("AssembleFromCode : %x\n", hRet);
  shader.DispHandle();
  shader.Release();
  
  hRet |= shader.CompileFromFile("test.fx", "basic_vs", "basic_ps");
  printf("CompileFromFile : %x\n", hRet);
  shader.DispHandle();
  shader.Release();
  
  hRet |= shader.CompileFromCode(shaderCode, 0, "basic_vs", "basic_ps");
  printf("CompileFromCode : %x\n", hRet);
  shader.DispHandle();
  shader.Release();
  
  printf("Please check Test result.\n");
}
