#include "common.h"

#include <process.h>

//----------------------------------------------------------------------------------------------
//
// Object
//
//----------------------------------------------------------------------------------------------
static LPDIRECTSOUND8       lpDS      = NULL;
static LPDIRECTSOUNDBUFFER8 lpDSDummy = NULL;
static DSCAPS               dwcaps;
static CRITICAL_SECTION     cs;
static HANDLE               hPollThread = NULL;
static unsigned int         dwPollThreadID  = NULL;

//؂ĈꎞIȃobt@Read悤ɂ
//I/F؂̖ʓ|
HRESULT FillSoundBuffer(LPDIRECTSOUNDBUFFER8 lpDSB8, void *src, BOOL bIsFill, DWORD dwLockOffset, DWORD dwLockSize, LPDWORD lpRes) {
  DWORD     dwLockSize1   = 0;
  DWORD     dwLockSize2   = 0;
  LPVOID    lpDest1       = NULL;
  LPVOID    lpDest2       = NULL;
  DWORD     dwRes1        = 0;
  DWORD     dwRes2        = 0;
  HRESULT   hRet          = S_FALSE;
  char     *lpSrc         = (char *)src;
  if(lpDSB8 && src) {
    hRet = lpDSB8->Lock( dwLockOffset, dwLockSize, &lpDest1, &dwLockSize1, &lpDest2, &dwLockSize2, 0);
    if( hRet != DS_OK) {
      *lpRes = 0;
      return hRet;
    }

    if(lpDest1) {
      if(bIsFill) {
        ::CopyMemory(lpDest1, lpSrc, dwLockSize1);
        lpSrc += dwLockSize1;
      } else {
        ZeroMemory(lpDest1, dwLockSize1);
        dwRes1 = dwLockSize1;
      }
    }

    if(lpDest2) {
      if(bIsFill) {
        ::CopyMemory(lpDest2, lpSrc, dwLockSize2);
        lpSrc += dwLockSize2;
      } else {
        ZeroMemory(lpDest2, dwLockSize2);
        dwRes2 = dwLockSize2;
      }
    }
    hRet = lpDSB8->Unlock(lpDest1, dwRes1, lpDest2, dwRes2);
  }

  if(lpRes) {
    *lpRes = dwRes1 + dwRes2;
  }
  return hRet;
}

//Create Sound Buffer
static LPDIRECTSOUNDBUFFER8 CreateSoundBuffer(WAVEFORMATEX *pwfx, DWORD dwSize, DWORD dwCaps) {
  LPDIRECTSOUNDBUFFER8 lpRet = NULL;
  if(lpDS && pwfx) {
    HRESULT               hRet          = E_FAIL;
    LPDIRECTSOUNDBUFFER   lpTempBuffer  = NULL;
    DSBUFFERDESC          dsbd          = {
      sizeof(DSBUFFERDESC), dwCaps, dwSize, 0, pwfx, GUID_NULL
    };
    
    hRet = lpDS->CreateSoundBuffer(&dsbd, &(lpTempBuffer), NULL);
    if(hRet != DS_OK) {
      return lpRet;
    }
    
    hRet = lpTempBuffer->QueryInterface( IID_IDirectSoundBuffer8, (void **)&(lpRet) );
    if(hRet != DS_OK) {
      lpTempBuffer->Release();
      return lpRet = NULL;
    }
    lpTempBuffer->Release();
  }
  return lpRet;
}

//------------------------------------------------------------------------------
// DirectSound with MMIO
//------------------------------------------------------------------------------
class WaveSound {
  //Sound Buffer
  LPDIRECTSOUNDBUFFER8  lpDSBuffer;
  WAVEFORMATEX          wfx;
  
  //MMIO Data
  HMMIO                 hmmio;
  MMCKINFO              mmckRiff;
  MMCKINFO              mmckData;
  
  //Buffer Size
  DWORD                 dwBufferSize;
  
  //Stream Buffer Data (polling)
  DWORD                 dwUnitSize;
  DWORD                 dwWritePos;
  BOOL                  bIsFront;
  BOOL                  bIsRead;
  
  BOOL                  bIsLooping;
  DWORD                 dwLoopPos;

  void Clear() {
    lpDSBuffer    = NULL;
    ::ZeroMemory(&wfx, sizeof(wfx));
    hmmio         = NULL;
    ::ZeroMemory(&mmckRiff, sizeof(mmckRiff));
    ::ZeroMemory(&mmckData, sizeof(mmckData));
    dwBufferSize  = 0;
    dwUnitSize    = 0;
    dwWritePos    = 0;
    bIsFront      = FALSE;
    bIsRead       = FALSE;
    bIsLooping    = FALSE;
    dwLoopPos     = 0;
  }
  
  void FillZeroSound(LPVOID lpdest, size_t n) {
    if(lpDSBuffer) {
      if(wfx.wBitsPerSample == 8) {
        ::FillMemory(lpdest, n, 128);
      } else {
        ::ZeroMemory(lpdest, n);
      }
    }
  }
  
  //Read Data use MMIO.
  HRESULT FillSoundBuffer(BOOL bIsFill, DWORD dwLockOffset, DWORD dwLockSize, LPDWORD lpRes) {
    DWORD     dwLockSize1     = 0;
    DWORD     dwLockSize2     = 0;
    LPVOID    lpDest1         = NULL;
    LPVOID    lpDest2         = NULL;
    DWORD     dwRes1          = 0;
    DWORD     dwRes2          = 0;
    DWORD     dwWrittenByte   = 0;
    DWORD     dwLockedSize    = 0;
    HRESULT   hRet            = S_FALSE;
    
    //Lock and Fill Data
    if(lpDSBuffer) {
      hRet = lpDSBuffer->Lock( dwLockOffset, dwLockSize, &lpDest1, &dwLockSize1, &lpDest2, &dwLockSize2, 0);
      if( hRet != DS_OK) {
        *lpRes = 0;
        return hRet;
      }

      //Store Locked Result Byte.
      dwLockedSize = dwLockSize1 + dwLockSize2;

      //Data1
      if(lpDest1) {
        if(bIsFill) {
          FillZeroSound(lpDest1, dwLockSize1);
          dwRes1 = mmioRead(hmmio, (HPSTR)lpDest1, dwLockSize1);
        } else {
          FillZeroSound(lpDest1, dwLockSize1);
          dwRes1 = dwLockSize1;
        }
      }

      //Data2
      if(lpDest2) {
        if(bIsFill) {
          FillZeroSound(lpDest2, dwLockSize2);
          dwRes2 = mmioRead(hmmio, (HPSTR)lpDest2, dwLockSize2);
        } else {
          FillZeroSound(lpDest2, dwLockSize2);
          dwRes2 = dwLockSize2;
        }
      }

      /*
      if(dwRes1 + dwRes2 < dwLockedSize ) {
        if(bIsLooping) {
          printf("seek loop position\n");
          //seek loop position
          mmioSeek(hmmio, dwLoopPos + mmckData.dwDataOffset, SEEK_SET);
          bIsFill = TRUE;
          
          //calc Write byte
          if(dwLockSize1) {
            dwLockSize1 -= dwRes1;
            char *temp = (char *)lpDest1;
            temp += dwRes1;
            lpDest1 = (LPVOID)temp;
          }
          
          if(dwLockSize2) {
            dwLockSize2 -= dwRes2;
            char *temp = (char *)lpDest2;
            temp += dwRes2;
            lpDest2 = (LPVOID)temp;
          }
        } else {
          //Done Write
          printf("Done Write\n");
          break;
        }
      }
      */
      
      //Update dwWritten Byte
      dwWrittenByte += dwRes1 + dwRes2;

      //Unlock
      hRet = lpDSBuffer->Unlock(lpDest1, dwRes1, lpDest2, dwRes2);
      
      //Store Written Byte
      if(lpRes) {
        *lpRes = dwWrittenByte;
      }
    }
    return hRet;
  }
  enum {
    SoundStreamSize   = 4,
    StreamBufferDivide = 2,
  };
  
public:
  enum {
    Effect,
    Stream,
  };
  
  WaveSound() {
    Clear();
  }

  ~WaveSound() {
    Close();
  }
  
  //Get Wave Data Size
  DWORD Size() {
    return dwBufferSize;
  }
  
  void Update() {
    DWORD dwRet = 0;
    DWORD dwPos = 0;
    
    //Is unit size enable, then this buffer is streaming.
    if(dwUnitSize) {
      DWORD dwBufferCenter = (dwBufferSize >> 1);
      lpDSBuffer->GetCurrentPosition(&dwPos, NULL);
      
      //printf("Front : Play Position : %d\n", dwPos);
      
      if(dwPos > dwBufferCenter && !bIsFront) {
        FillSoundBuffer(bIsRead, dwWritePos, dwUnitSize, &dwRet);
        if(dwRet < dwUnitSize && bIsLooping) {
          DWORD temp = dwUnitSize - dwRet;
          mmioSeek(hmmio, dwLoopPos + mmckData.dwDataOffset, SEEK_SET);
          FillSoundBuffer(bIsRead, dwWritePos + dwRet, temp, NULL);
        }
        dwWritePos = dwUnitSize;
        
        //printf("Next Write Position : %d\n", dwWritePos);
        
        bIsFront = TRUE;
        if(dwRet <= 0) {
          bIsRead = FALSE;
        }
      }

      if(dwPos < dwBufferCenter && bIsFront) {
        FillSoundBuffer(bIsRead, dwWritePos, dwUnitSize, &dwRet);
        if(dwRet < dwUnitSize && bIsLooping) {
          DWORD temp = dwUnitSize - dwRet;
          mmioSeek(hmmio, dwLoopPos + mmckData.dwDataOffset, SEEK_SET);
          FillSoundBuffer(bIsRead, dwWritePos + dwRet, temp, NULL);
        }
        dwWritePos = 0;
        
        //printf("Next Write Position : %d\n", dwWritePos);
        
        bIsFront = FALSE;
        if(dwRet <= 0) {
          bIsRead = FALSE;
        }
      }
    }
  }
  
  HRESULT SetLoop(BOOL b) {
    bIsLooping = b;
    return S_OK;
  }

  HRESULT Play() {
    HRESULT hRet = E_FAIL;
    if(lpDSBuffer) {
      if(!dwUnitSize) {
        lpDSBuffer->SetCurrentPosition(0);
        hRet = lpDSBuffer->Play(0, 0, 0);
      } else {
        ::EnterCriticalSection(&cs);
        hRet = lpDSBuffer->Play(0, 0, DSBPLAY_LOOPING);
        ::LeaveCriticalSection(&cs);
      }
    }
    return hRet;
  }
  
  HRESULT Stop() {
    if(lpDSBuffer) {
      return lpDSBuffer->Stop();
    }
    return E_FAIL;
  }
  
  HRESULT OpenFromFile(LPCSTR szName, DWORD type) {
    HRESULT   hRet = E_FAIL;
    MMCKINFO  mmckCheck;
    DWORD dwRet;
    DWORD dwCaps =  DSBCAPS_LOCSOFTWARE |
                    DSBCAPS_CTRLFREQUENCY |
                    DSBCAPS_CTRLPAN |
                    DSBCAPS_CTRLVOLUME |
                    DSBCAPS_GLOBALFOCUS |
                    DSBCAPS_GETCURRENTPOSITION2;
    if(type == Stream) ::EnterCriticalSection(&cs);
    
    Close();
    ZeroMemory(&mmckCheck, sizeof(mmckCheck));
    
    //Throw Error 
    do {
      if(szName == NULL) break;
      
      //
      // M M I O  F i l e  O p e n 
      //
      hmmio = ::mmioOpen((LPSTR)szName, NULL, MMIO_READ | MMIO_ALLOCBUF);
      if(hmmio == NULL) {
        break;
      }
      
      //Check Wave File
      mmioDescend(hmmio, &mmckRiff, NULL, 0);
      if(mmckRiff.ckid != FOURCC_RIFF || mmckRiff.fccType != mmioFOURCC('W','A','V','E') ) {
        Close();
        break;
      }
      
      mmckCheck.ckid = mmioFOURCC('f', 'm', 't', ' ');  
      if(mmioDescend(hmmio, &mmckCheck, &mmckRiff, MMIO_FINDCHUNK) != MMSYSERR_NOERROR) {
        Close();
        break;
      }

      //Read WAVEFORMATEX
      if(mmioRead(hmmio, (HPSTR)&wfx, sizeof(wfx) ) != sizeof(wfx)) {
        Close();
        break;
      }

      //Check PCM tag.
      if(wfx.wFormatTag != WAVE_FORMAT_PCM) {
        Close();
        break;
      }
      
      //descend data chunk.
      mmioSeek(hmmio, mmckRiff.dwDataOffset + sizeof(FOURCC), SEEK_SET);
      mmckData.ckid = mmioFOURCC('d', 'a', 't', 'a');  
      if(mmioDescend(hmmio, &mmckData, &mmckRiff, MMIO_FINDCHUNK) != MMSYSERR_NOERROR) {
        Close();
        break;
      }
      //MMIO Open done.

      //Effect or Stream
      if(type == Stream) {
        dwUnitSize   = wfx.nSamplesPerSec * SoundStreamSize * wfx.nBlockAlign;
        dwUnitSize  /= StreamBufferDivide;
        dwUnitSize  -= dwUnitSize % wfx.nBlockAlign;
        dwBufferSize = dwUnitSize * StreamBufferDivide;
      } else {
        dwBufferSize = mmckData.cksize;
        dwUnitSize = 0;
        dwWritePos = 0;
        bIsFront = FALSE;
      }
      
      lpDSBuffer = CreateSoundBuffer(&wfx, dwBufferSize, dwCaps);
      if(!lpDSBuffer) {
        Close();
        break;
      }
      
      hRet =  S_OK;
    } while(0);
    
    if(hRet == S_OK) {
      bIsRead = TRUE;
      hRet = FillSoundBuffer(bIsRead, 0, dwBufferSize, &dwRet);
    }
    
    if(type == Stream) ::LeaveCriticalSection(&cs);
    
    return hRet;
  }


  HRESULT Close() {
    HRESULT hRet = S_FALSE;
    if(dwUnitSize) ::EnterCriticalSection(&cs);
    Stop();
    if(hmmio) {
      ::mmioClose(hmmio, 0);
    }
    RELEASE(lpDSBuffer);
    if(dwUnitSize) ::LeaveCriticalSection(&cs);
    Clear();
    return hRet;
  }
};

static WaveSound wsound[SoundMax];

//----------------------------------------------------------------------------------------------
//
// Function
//
//----------------------------------------------------------------------------------------------
DWORD GetMaxSampleRate() {
  return dwcaps.dwMaxSecondarySampleRate;
}

DWORD GetChannels() {
  return (dwcaps.dwFlags  & DSCAPS_PRIMARYSTEREO)  ?  2 : 1;
}

DWORD GetBitPerRate() {
  return (dwcaps.dwFlags  & DSCAPS_SECONDARY16BIT) ? 16 : 8;
}

HRESULT TermDirectSound() {
  HRESULT hRet = E_FAIL;
  
  //Stop all sound
  for(int i = 0 ; i < SoundMax; i++) {
    wsound[i].Stop();
  }
  
  //Close all sound.
  for(int i = 0 ; i < SoundMax; i++) {
    wsound[i].Close();
  }
  
  //Termnate Thread
  if(hPollThread) {
    ::PostThreadMessage(dwPollThreadID, WM_QUIT, 0, 0);
    ::WaitForSingleObject(hPollThread, INFINITE);
    ::CloseHandle(hPollThread);
    hPollThread = NULL;
  }
  
  
  RELEASE(lpDSDummy);
  RELEASE(lpDS);
  ::ZeroMemory(&dwcaps, sizeof(dwcaps));
  ::DeleteCriticalSection(&cs);
  ::CoUninitialize();
  
  hRet = S_OK;
  
  return hRet;
}

static HRESULT PlayDummyBuffer(DWORD dwChannels, DWORD dwBitRate, DWORD dwFreq) {
  HRESULT hRet = E_FAIL;
  if(lpDSDummy == NULL) {
    DSBUFFERDESC  dsbd;
    WAVEFORMATEX  wfx;
    DWORD       dwCaps =
              DSBCAPS_LOCSOFTWARE | 
              DSBCAPS_CTRLFREQUENCY | 
              DSBCAPS_CTRLPAN | 
              DSBCAPS_CTRLVOLUME | 
              DSBCAPS_CTRLPOSITIONNOTIFY | 
              DSBCAPS_GLOBALFOCUS | 
              DSBCAPS_GETCURRENTPOSITION2;
    
    //WAVEFORMAT^O쐬
    ZeroMemory(&wfx, sizeof(wfx));
    wfx.wFormatTag      = WAVE_FORMAT_PCM;
    wfx.nChannels       = (WORD)dwChannels;
    wfx.nSamplesPerSec  = dwFreq;
    wfx.wBitsPerSample  = (WORD)dwBitRate;
    wfx.nBlockAlign     = wfx.nChannels * wfx.wBitsPerSample / 8;
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
    wfx.cbSize          = sizeof(WAVEFORMATEX);

    ZeroMemory(&dsbd, sizeof(dsbd));
    dsbd.dwSize          = sizeof(DSBUFFERDESC);
    dsbd.dwBufferBytes   = wfx.nSamplesPerSec * 3 * wfx.nBlockAlign;
    dsbd.dwFlags         = dwCaps;
    dsbd.lpwfxFormat     = &wfx;
    dsbd.guid3DAlgorithm = DS3DALG_DEFAULT;
    LPDIRECTSOUNDBUFFER  lpTempBuffer;
    hRet = lpDS->CreateSoundBuffer(&dsbd, &(lpTempBuffer), NULL);
    if(hRet != DS_OK) {
      return hRet;
    }
    
    hRet = lpTempBuffer->QueryInterface( IID_IDirectSoundBuffer8, (void **)&(lpDSDummy) );
    if(hRet != DS_OK) {
      lpTempBuffer->Release();
      return hRet;
    }
    RELEASE(lpTempBuffer);
    
    char *buf = new char[dsbd.dwBufferBytes];
    if(buf) {
      ::ZeroMemory(buf, dsbd.dwBufferBytes);
      FillSoundBuffer(lpDSDummy, buf, TRUE, 0, dsbd.dwBufferBytes, NULL);
      delete [] buf;
      
      hRet = lpDSDummy->Play(0, 0, DSBPLAY_LOOPING);
    } else {
      hRet = E_OUTOFMEMORY;
    }
  }
  
  return hRet;
}


//Update Sound Thread
static unsigned int WINAPI ThreadUpdateSound(LPVOID lp) {
  MSG msg;
  BOOL bDone = FALSE;
  while(!bDone) {
    
    //update sound
    ::EnterCriticalSection(&cs);
    for(int i = 0 ; i < SoundMax; i++) {
      wsound[i].Update();
    }
    ::LeaveCriticalSection(&cs);
    
    //if 
    while( ::PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {
      if(msg.message == WM_QUIT) {
        bDone = TRUE;
        break;
      }
    }
    if(!bDone) ::Sleep(PollThreadSleep);
  }
  return 0;
}


HRESULT InitDirectSound() {
  HRESULT hRet = E_FAIL;
  
  if(lpDS != NULL) {
    return hRet;
  }
  ::CoInitialize(NULL);
  hRet = DirectSoundCreate8(NULL, &lpDS, NULL);
  if( hRet != DS_OK ) {
    return hRet;
  }
  
  hRet = lpDS->SetCooperativeLevel(GetWindowHandle(), DSSCL_NORMAL);
  if(hRet != DS_OK) {
    TermDirectSound();
    return hRet;
  }
  
  ZeroMemory(&dwcaps, sizeof(dwcaps));
  dwcaps.dwSize = sizeof(dwcaps);
  hRet = lpDS->GetCaps(&dwcaps);
  if(hRet == DS_OK) {
    DWORD dwChannel = (dwcaps.dwFlags  & DSCAPS_PRIMARYSTEREO)  ?  2 : 1;
    DWORD dwBitRate = (dwcaps.dwFlags  & DSCAPS_SECONDARY16BIT) ? 16 : 8;
    hRet = PlayDummyBuffer(dwChannel, dwBitRate, dwcaps.dwMaxSecondarySampleRate );
  }
  
  //Create Update Thread
  ::InitializeCriticalSection(&cs);
  hPollThread = (HANDLE)_beginthreadex(NULL, 0, ThreadUpdateSound, NULL, 0, &dwPollThreadID);
  
  return hRet;
}

static BOOL CheckSoundMax(DWORD num) {
  if(SoundMax < num) return TRUE;
  return FALSE;
}

static HRESULT LoadSound(DWORD num, LPCSTR szName, DWORD type) {
  HRESULT hRet = E_FAIL;
  WaveSound *lpsound;
  if(!CheckSoundMax(num)) {
    lpsound = &(wsound[num]);
    if(!lpsound->Size()) {
      hRet = lpsound->OpenFromFile(szName, type);
    }
  }
  return hRet;
}

HRESULT LoadSoundFromFile(DWORD num, LPCSTR szName) {
  return LoadSound(num, szName, WaveSound::Effect);
}

HRESULT LoadStreamSoundFromFile(DWORD num, LPCSTR szName) {
  return LoadSound(num, szName, WaveSound::Stream);
}

HRESULT SetLoopSound(DWORD num, BOOL b) {
  HRESULT hRet = E_FAIL;
  WaveSound *lpsound;
  if(!CheckSoundMax(num)) {
    lpsound = &(wsound[num]);
    hRet = lpsound->SetLoop(b);
  }
  return hRet;
}


HRESULT PlaySound(DWORD num) {
  HRESULT hRet = E_FAIL;
  WaveSound *lpsound;
  if(!CheckSoundMax(num)) {
    lpsound = &(wsound[num]);
    
    SetLoopSound(num, TRUE);
    hRet = lpsound->Play();
  }
  return hRet;
}


HRESULT StopSound(DWORD num) {
  HRESULT hRet = E_FAIL;
  WaveSound *lpsound;
  if(!CheckSoundMax(num)) {
    lpsound = &(wsound[num]);
    hRet = lpsound->Stop();
  }
  return hRet;
}

HRESULT   UnloadSound(DWORD num) {
  HRESULT hRet = E_FAIL;
  WaveSound *lpsound;
  if(!CheckSoundMax(num)) {
    lpsound = &(wsound[num]);
    hRet = lpsound->Close();
  }
  return hRet;
}


